@@ -80,14 +80,6 @@ child_free(void *data)
     g_slice_free(Child, child);
 }
 
-/* If we set JSCLASS_CONSTRUCT_PROTOTYPE flag, then this is called on
- * the prototype in addition to on each instance. When called on the
- * prototype, "obj" is the prototype, and "retval" is the prototype
- * also, but can be replaced with another object to use instead as the
- * prototype. If we don't set JSCLASS_CONSTRUCT_PROTOTYPE we can
- * identify the prototype as an object of our class with NULL private
- * data.
- */
 GJS_NATIVE_CONSTRUCTOR_DECLARE(keep_alive)
 {
     GJS_NATIVE_CONSTRUCTOR_VARIABLES(keep_alive)
@@ -186,9 +178,7 @@ keep_alive_trace(JSTracer *tracer,
  */
 static struct JSClass gjs_keep_alive_class = {
     "__private_GjsKeepAlive", /* means "new __private_GjsKeepAlive()" works */
-#ifdef HAVE_JS_CLASS_TRACE
     JSCLASS_MARK_IS_TRACE | /* TraceOp not MarkOp */
-#endif
     JSCLASS_HAS_PRIVATE,
     JS_PropertyStub,
     JS_PropertyStub,
@@ -204,11 +194,7 @@ static struct JSClass gjs_keep_alive_class = {
     NULL,
     NULL,
     NULL,
-#ifdef HAVE_JS_CLASS_TRACE
     JS_CLASS_TRACE(keep_alive_trace),
-#else
-    keep_alive_trace,
-#endif
     NULL
 };
 
